/*
 *	mp3tag.c  -- MP3 TAG Editor.
 *
 *	Copyright (C) 1998  Marcin Baranowski (minek@kis.p.lodz.pl).
 *      [ send comments and bug reports to minek@kis.p.lodz.pl]
 *
 *    This is very ALPHA code (may be unstable in some cases)
 * 
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	
 *  
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "mp3tag.h"

const char *VERSION="0.1b"; 
  const struct Genre genre[116] = 
{
	{	"Acid", 			0x22	},
	{	"Acid Jazz",		0x4A	},
	{	"Acid Punk",		0x49	},
	{	"Acoustic",			0x63	},
	{	"Alternative",		0x14	},
	{	"Alt. Rock",		0x28	},
	{	"Ambient",			0x1A	},
	{	"Avantgarde",		0x5A	},
	{	"Bass",				0x29	},
	{	"Bebob",			0x55	},
	{	"Big Band",			0x60	},
	{	"Bluegrass",		0x59	},
	{	"Blues",			0x00	},
	{	"Booty Bass",		0x6B	},
	{	"Cabaret",			0x41	},
	{	"Celtic",			0x58	},
	{	"Chamber Music",	0x68	},
	{	"Chanson",			0x66	},
	{	"Chorus",			0x61	},
	{	"Christian Rap",	0x3D	},
	{	"Classical",		0x20	},
	{	"Classic Rock",		0x01	},
	{	"Club",				0x70	},
	{	"Comedy",			0x39	},
	{	"Country",			0x02	},
	{	"Cult",				0x3A	},
	{	"Darkwave",			0x32	},
	{	"Dance", 			0x03	},
	{	"Death Metal",		0x16	},
	{	"Disco",			0x04	},
	{	"Dream",			0x37	},
	{	"Easy Listening",	0x62	},
	{	"Electronic",		0x34	},
	{	"Ethnic",			0x30	},
	{	"Eurodance",		0x36	},
	{	"Euro-Techno",		0x19	},
	{	"Fast-Fusion",		0x54	},
	{	"Folk",				0x50	},
	{	"Folk/Rock",		0x51	},
	{	"Folklore",			0x73	},
	{	"Funk",				0x05	},
	{	"Fusion",			0x1E	},
	{	"Game",				0x24	},
	{	"Gangsta",			0x3B	},
	{	"Gothic",			0x31	},
	{	"Gothic Rock",		0x5B	},
	{	"Gospel",			0x26	},
	{	"Grunge",			0x06	},
	{	"Hard Rock",		0x4F	},
	{	"Hip-Hop",			0x07	},
	{	"House", 			0x23	},
	{	"Humour",			0x64	},
	{	"Industrial",		0x13	},
	{	"Instrumental",		0x21	},
	{	"Instrumental Pop",	0x2E	},
	{	"Instrumental Rock",0x2F	},
	{	"Jazz", 			0x08	},
	{	"Jazz+Funk",		0x1D	},
	{	"Jungle",			0x3F	},
	{	"Latin",			0x56	},
	{	"Lo-Fi",			0x47	},
	{	"Meditative",		0x2D	},
	{	"Metal",			0x09	},
	{	"Musical",			0x4D	},
	{	"National Folk",	0x52	},
	{	"Native American",	0x40	},
	{	"New Age",			0x0A	},
	{	"New Wave",			0x42	},
	{	"Noise",			0x27	},
	{	"Oldies",			0x0B	},
	{	"Opera",			0x67	},
	{	"Other",			0x0C	},
	{	"Polka",			0x4B	},
	{	"Pop",				0x0D	},
	{	"Pop/Funk",			0x3E	},
	{	"Pop-Folk",			0x35	},
	{	"Porn Groove",		0x6D	},
	{	"Primus",			0x6C	},
	{	"Progressive Rock",	0x5C	},
	{	"Punk",				0x2B	},
	{	"Pranks",			0x17	},
	{	"Psychadelic",		0x43	},
	{	"Psychedelic Rock",	0x5D	},
	{	"Rap",				0x0F	},
	{	"Rave",				0x44	},
	{	"Reggae",			0x10	},
	{	"Revival",			0x57	},
	{	"Retro",			0x4C	},
	{	"Rock",				0x11	},
	{	"Rock & Roll",		0x4E	},
	{	"R&B",				0x0E	},
	{	"Samba", 			0x72	},
	{	"Satire",			0x6E	},
	{	"Showtunes",		0x45	},
	{	"Ska",				0x15	},
	{	"Slow Jam",			0x6F	},
	{	"Slow Rock",		0x5F	},
	{	"Sonata",			0x69	},
	{	"Soul",				0x2A	},
	{	"Soundtrack",		0x18	},
	{	"Sound Clip",		0x25	},
	{	"Southern Rock",	0x38	},
	{	"Space",			0x2C	},
	{	"Speech",			0x65	},
	{	"Swing",			0x53	},
	{	"Symphony",			0x6A	},
	{	"Symphonic Rock",	0x5E	},
	{	"Tango",			0x71	},
	{	"Techno",			0x12	},
	{	"Techno-Industrial",0x33	},
	{	"Top 40",			0x3C	},
	{	"Trailer",			0x46	},
	{	"Tribal",			0x48	},
	{	"Trance",			0x1F	},
	{	"Trip-Hop",			0x1B	},
	{	"Vocal",			0x1C	}
};
 
 
 
 
 
 
 
struct TAGmp3 mp3tag; 
FILE *mp3file;
const char *mp3name;
const char *pname;
Boolean IsX;
pid_t p;
char checktag( )
{
	char tag[3];
	fseek( mp3file, -128, SEEK_END );
	fscanf( mp3file, "%3s", tag );
	if ( strcmp( tag, TAG ) )
		return 0;
	return 1;
}

int readtag()
{
	if ( ! checktag( ) ) 
	{
		printf( "File %s is without tag\n\r", mp3name );
		
		if(IsX) return(-1);
				 else exit(0);
	}
	
	fscanf( mp3file, "%30c", mp3tag.title );
	fscanf( mp3file, "%30c", mp3tag.artist );
	fscanf( mp3file, "%30c", mp3tag.album );
	fscanf( mp3file, "%4c", mp3tag.year );
	fscanf( mp3file, "%30c", mp3tag.comment );
	fscanf( mp3file, "%c", &mp3tag.genre );
	return(0);
}

void readtag_frag( FILE *tagfile, char *frag, char count )
{	
	char ch, i = 0;
	while( ( ch = fgetc( tagfile ) ) != '\n' )
	{
		if ( ( i > count ) || ( ch == '\n' ) ) 
			break;
		frag[i++] = ch;
	}
}

char genread( char *gen )
{
	char i;
	for ( i = 0; i < GMAX; i++ )
	{
		if ( ! ( strcmp( gen, genre[i].genre ) ) )
			return genre[i].ch;
	}
	return 0xFF;
}

void fgenread( FILE *tagfile )
{
	char frag[18], ch, i = 0;
	while( ( ch = fgetc( tagfile ) ) != '\n' )
	{
		if ( ( i > 17 ) || ( ch == '\n' ) ) 
			break;
		frag[i++] = ch;
	}
	frag[i++] = '\0';
	mp3tag.genre = genread( frag );
}		

const char *genwrite( char gen )
{
	char i;
	for ( i = 0; i < GMAX; i++ )
	{
		if ( genre[i].ch == gen )
			return genre[i].genre;
	}
	return " ";
}

int readtagfile( char *filename )
{
	FILE *tagfile;

	if ( !( tagfile = fopen( filename, "r" ) ) )
	{
		fprintf( stderr, "Error: can't open file %s\n\r", filename );
		
		if(IsX) return(-1);
				 else exit(0);
	}

	strcpy( mp3tag.tag, TAG );
	readtag_frag( tagfile, mp3tag.title, 30 );
	readtag_frag( tagfile, mp3tag.artist, 30 );
	readtag_frag( tagfile, mp3tag.album, 30 );
	readtag_frag( tagfile, mp3tag.year, 4 );
	fgenread( tagfile );
	readtag_frag( tagfile, mp3tag.comment, 30 );
	return(0);
	
}

void print_frg( char *frag, char count )
{
	char i;
	for ( i = 0; ( i < count ) && frag[i]; i++ )
		putchar( frag[i] );
}

char *print_frag( char *frag, char count )
{
	char *buf, i;
	buf = malloc( count );
	for ( i = 0; i < count; i++ )
		buf[i] = frag[i];
	return buf;
}

void writetag_frag( const char *tag, char count )
{
	char i; //, *buf;
	for ( i = 0; i < count; i++ )
		fputc( tag[i] ? tag[i] : 0x21, mp3file );
}

void savetag( )
{
	
	if ( checktag( ) )
		fseek( mp3file, -128, SEEK_END );
	else 
		fseek( mp3file, 0, SEEK_END );
/*
	printf( "%ld\n", ftell( mp3file ) );
	fwrite( (void *)ftell( mp3file ), 3, writetag_frag( TAG, 3 ), mp3file ); 
	fwrite( (void *) ftell( mp3file ), 30, writetag_frag( mp3tag.title, 30 ), mp3file );
	fwrite( (void *) ftell( mp3file ), 30, writetag_frag( mp3tag.artist, 30 ), mp3file );
	fwrite( (void *) ftell( mp3file ), 30, writetag_frag( mp3tag.album, 30 ), mp3file );
	fwrite( (void *) ftell( mp3file ), 4, writetag_frag( mp3tag.year, 4 ), mp3file );
	fwrite( (void *) ftell( mp3file ), 30, writetag_frag( mp3tag.comment, 30 ), mp3file );
	fwrite( (void *) ftell( mp3file ), 1, writetag_frag( genwrite( mp3tag.genre ), 1 ), mp3file );
*/
	writetag_frag( TAG, 3 ); 
	writetag_frag( mp3tag.title, 30 );
	writetag_frag( mp3tag.artist, 30 );
	writetag_frag( mp3tag.album, 30 );
	writetag_frag( mp3tag.year, 4 );
	writetag_frag( mp3tag.comment, 30 );
	writetag_frag( &mp3tag.genre, 1 );

}

void printtag( char user )
{
	printf( "%s", user ? "Title: " : "" );
	print_frg( mp3tag.title, 30 );
	printf( "\n" );
	printf( "%s", user ? "Artist: " : "" ); 
					print_frg( mp3tag.artist, 30  );
	printf( "\n" );
	
	printf( "%s", user ? "Album: " : "" ); 
					print_frg( mp3tag.album, 30 );
	printf( "\n" );
	
	printf( "%s", user ? "Year: " : "" );
					print_frg( mp3tag.year, 4  );
	printf( "\n" );
	
	printf( "%s%s\n", user ? "Type: " : "",
					genwrite( mp3tag.genre ) );
	
	printf( "%s", user ? "Comment: " : "" );
					print_frg( mp3tag.comment, 30 );
	printf( "\n" );
}

void striptag( )
{
	char i;
	if ( !( checktag( ) ) )
	{
		printf( "File %s hasn't tag - nothing to cut.\n\r", 
						mp3name );
		return;
	}
	else 
		fseek( mp3file, -128, SEEK_END );
	for ( i = 0; i < 127; i ++ )
		fputc( EOF, mp3file );
}	

void usage( )
{
	
	printf("MP3 TAG Editor, Marcin Baranowski (minek@kis.p.lodz.pl)\n");
	#ifdef xver
	printf("X11 interface, Marek Blaszkowski (mb@kis.p.lodz.pl)\n");
	#endif
	printf( "Usage: \n" );
	printf( "%s [args] \n", pname );
	printf( "\t-p <file.mp3> - print tag from file (if any exist).\n " );
	printf( "\t-d <file.mp3> - print tag in form accepted by this editor.\n " );
	printf( "\t-a <file.mp3> <file_with_tag> - insert to  file <*.mp3>\n\t\ttag from <file_with_tag>. \n " );
	#ifdef xver
	printf( "\t-x X11 interface\n");
	#endif
	printf("\t-v show version\n");
	exit( 0 );
}

int openmp3file( const char *filename )
{
	if ( !filename )
	{
		fprintf( stderr, "Give me a file .mp3.\n\r" );
		exit( 1 );
	}
	mp3name = filename;

	if ( !( mp3file = fopen( mp3name, "rb+" ) ) )
	{
	if(errno==2){fprintf( stderr, "Error: can't open file %s %d!\n\r", mp3name );
				return(-1);
				}
	if (errno==13){
				 if ( !( mp3file = fopen( mp3name, "r" ) ) )
					{
					fprintf( stderr, "Error: can't open file %s %d!\n\r", mp3name );
					return(-1);
					}
				 return(-2);	
					}
					
	}
	return(0);
/*	printf("%s \n",mp3name); */
}

void parsecommand( int argc, char **argv )

{
	register char *s;
	char *tmparg;
	char x;
	tmparg=argv[0];
	while( --argc > 0 && (*++argv)[0] == '-' )
		for ( s = argv[0]+1; *s != '\0'; s++ )
			switch( *s )
			{
			case 'p' :
				--argc;
				openmp3file( *++argv );
				readtag();
				printtag( 1 );
				fclose( mp3file );
				break;
			case 'd' :
				--argc;
				openmp3file( *++argv );
				readtag();
				printtag( 0 );
				fclose( mp3file );
				break;
			case 'a' :
				--argc;
				openmp3file( *++argv );
				--argc;
				readtagfile( *++argv );
				savetag();
				fclose( mp3file );
				break;
			#ifdef xver	
			case 'x' :
				argv[0]=tmparg;
				
				startupx(argc,argv);
				
				break;	
			#endif	
			case 'v':
				 printf("MP3 TAG Editor version %s\n",VERSION);
				 break;	
			default :
				break;
			}
}

int main( int argc, char **argv )
{	
IsX=FALSE;
	pname = argv[0]; 
	if ( argc < 2 )
		usage();

	parsecommand( argc, argv );
	
	return 0;
}
